﻿#include "pch.h"
#include "Sample3DSceneRenderer.h"

#include "..\Common\DirectXHelper.h"
#include <ppltasks.h>
#include <synchapi.h>

using namespace PathTransform;

using namespace Concurrency;
using namespace DirectX;
using namespace Microsoft::WRL;
using namespace Windows::Foundation;
using namespace Windows::Storage;

// Indices into the application state map.
Platform::String^ AngleKey = "Angle";
Platform::String^ TrackingKey = "Tracking";

constexpr int TestComputeNumBufferItems = 128;

struct SVec3DF
{
	float x;
	float y;
	float z;
};

struct SVec2DF
{
	float x;
	float y;
};

static const SVec3DF g_pathVertices[] =
{
	{ 200.000000f, 624.000000f, 0.000000f },
	{ 310.456940f, 624.000000f, 0.000000f },
	{ 400.000000f, 713.543030f, 0.000000f },
	{ 400.000000f, 824.000000f, 0.000000f },
	{ 400.000000f, 824.000000f, 0.000000f },
	{ 400.000000f, 934.456970f, 0.000000f },
	{ 310.456940f, 1024.000000f, 0.000000f },
	{ 200.000000f, 1024.000000f, 0.000000f },
	{ 0.000000f, 824.000000f, 0.000000f },
	{ 0.000000f, 934.456970f, 0.000000f },
	{ 89.543053f, 1024.000000f, 0.000000f },
	{ 200.000000f, 1024.000000f, 0.000000f },
	{ 200.000000f, 624.000000f, 0.000000f },
	{ 89.543053f, 624.000000f, 0.000000f },
	{ -0.000000f, 713.543030f, 0.000000f },
	{ 0.000000f, 824.000000f, 0.000000f }
};

constexpr UINT64 Align(const UINT64 value, const UINT64 alignment)
{
	return (value + alignment - 1) & ~(alignment - 1);
}

// Loads vertex and pixel shaders from files and instantiates the cube geometry.
Sample3DSceneRenderer::Sample3DSceneRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) :
	m_loadingComplete(false),
	m_radiansPerSecond(XM_PIDIV4),	// rotate 45 degrees per second
	m_angle(0),
	m_tracking(false),
	m_mappedConstantBuffer(nullptr),
	m_deviceResources(deviceResources),
	m_transformedVertices(false),
	m_testedCompute(false)
{
	auto d3dDevice = m_deviceResources->GetD3DDevice();

	ComPtr<ID3D12InfoQueue> info;

	if(SUCCEEDED(d3dDevice->QueryInterface<ID3D12InfoQueue>(&info)))
	{
		std::unique_ptr<D3D12_MESSAGE_ID[]> msgIDs = std::make_unique<D3D12_MESSAGE_ID[]>(D3D12_MESSAGE_ID_D3D12_MESSAGES_END);

		for(int i = 0; i < D3D12_MESSAGE_ID_D3D12_MESSAGES_END; ++i)
		{
			msgIDs[i] = (D3D12_MESSAGE_ID)i;
		}

		D3D12_MESSAGE_CATEGORY debug_io_categories[] =
		{
			D3D12_MESSAGE_CATEGORY_MISCELLANEOUS,
			D3D12_MESSAGE_CATEGORY_INITIALIZATION,
			D3D12_MESSAGE_CATEGORY_CLEANUP,
			D3D12_MESSAGE_CATEGORY_COMPILATION,
			D3D12_MESSAGE_CATEGORY_STATE_CREATION,
			D3D12_MESSAGE_CATEGORY_STATE_SETTING,
			D3D12_MESSAGE_CATEGORY_STATE_GETTING,
			D3D12_MESSAGE_CATEGORY_RESOURCE_MANIPULATION,
			D3D12_MESSAGE_CATEGORY_EXECUTION,
			D3D12_MESSAGE_CATEGORY_SHADER
		};

		std::vector<D3D12_MESSAGE_SEVERITY> debug_io_severities;

		debug_io_severities.push_back(D3D12_MESSAGE_SEVERITY_CORRUPTION);

		//if((debugVerbosity & TRenderingSystemDebugVerbosity::Information) == TRenderingSystemDebugVerbosity::Information)
		{
			debug_io_severities.push_back(D3D12_MESSAGE_SEVERITY_INFO);
			debug_io_severities.push_back(D3D12_MESSAGE_SEVERITY_MESSAGE);
		}

		//if((debugVerbosity & TRenderingSystemDebugVerbosity::Errors) == TRenderingSystemDebugVerbosity::Errors)
		{
			debug_io_severities.push_back(D3D12_MESSAGE_SEVERITY_ERROR);
		}

		/*if((debugVerbosity & TRenderingSystemDebugVerbosity::Warnings) == TRenderingSystemDebugVerbosity::Warnings
			|| (debugVerbosity & TRenderingSystemDebugVerbosity::PerformanceWarnings) == TRenderingSystemDebugVerbosity::PerformanceWarnings)*/
		{
			debug_io_severities.push_back(D3D12_MESSAGE_SEVERITY_WARNING);
		}

		D3D12_INFO_QUEUE_FILTER filters;
		ZeroMemory(&filters, sizeof(filters));

		filters.AllowList.NumCategories = _countof(debug_io_categories);
		filters.AllowList.pCategoryList = debug_io_categories;
		filters.AllowList.NumSeverities = static_cast<UINT>(debug_io_severities.size());
		filters.AllowList.pSeverityList = debug_io_severities.data();
		filters.AllowList.NumIDs = (UINT)D3D12_MESSAGE_ID_D3D12_MESSAGES_END;
		filters.AllowList.pIDList = msgIDs.get();

		info->PushRetrievalFilter(&filters);
		info->PushStorageFilter(&filters);
	}

	LoadState();
	ZeroMemory(&m_constantBufferData, sizeof(m_constantBufferData));

	CreateDeviceDependentResources();
	CreateWindowSizeDependentResources();
}

Sample3DSceneRenderer::~Sample3DSceneRenderer()
{
	m_constantBuffer->Unmap(0, nullptr);
	m_mappedConstantBuffer = nullptr;
}

void Sample3DSceneRenderer::CreateDeviceDependentResources()
{
	auto d3dDevice = m_deviceResources->GetD3DDevice();

	// Create a root signature with a single constant buffer slot.
	{
		CD3DX12_DESCRIPTOR_RANGE range;
		CD3DX12_ROOT_PARAMETER parameter;

		range.Init(D3D12_DESCRIPTOR_RANGE_TYPE_CBV, 1, 0);
		parameter.InitAsDescriptorTable(1, &range, D3D12_SHADER_VISIBILITY_VERTEX);

		D3D12_ROOT_SIGNATURE_FLAGS rootSignatureFlags =
			D3D12_ROOT_SIGNATURE_FLAG_ALLOW_INPUT_ASSEMBLER_INPUT_LAYOUT | // Only the input assembler stage needs access to the constant buffer.
			D3D12_ROOT_SIGNATURE_FLAG_DENY_DOMAIN_SHADER_ROOT_ACCESS |
			D3D12_ROOT_SIGNATURE_FLAG_DENY_GEOMETRY_SHADER_ROOT_ACCESS |
			D3D12_ROOT_SIGNATURE_FLAG_DENY_HULL_SHADER_ROOT_ACCESS |
			D3D12_ROOT_SIGNATURE_FLAG_DENY_PIXEL_SHADER_ROOT_ACCESS;

		CD3DX12_ROOT_SIGNATURE_DESC descRootSignature;
		descRootSignature.Init(1, &parameter, 0, nullptr, rootSignatureFlags);

		ComPtr<ID3DBlob> pSignature;
		ComPtr<ID3DBlob> pError;
		DX::ThrowIfFailed(D3D12SerializeRootSignature(&descRootSignature, D3D_ROOT_SIGNATURE_VERSION_1, pSignature.GetAddressOf(), pError.GetAddressOf()));
		DX::ThrowIfFailed(d3dDevice->CreateRootSignature(0, pSignature->GetBufferPointer(), pSignature->GetBufferSize(), IID_PPV_ARGS(&m_rootSignature)));
        NAME_D3D12_OBJECT(m_rootSignature);
	}

	// Load shaders asynchronously.
	auto createVSTask = DX::ReadDataAsync(L"SampleVertexShader.cso").then([this](std::vector<byte>& fileData) {
		m_vertexShader = fileData;
	});

	auto createPSTask = DX::ReadDataAsync(L"SamplePixelShader.cso").then([this](std::vector<byte>& fileData) {
		m_pixelShader = fileData;
	});

	auto createPathTransformCSTask = DX::ReadDataAsync(L"ui_pathTransform.cso").then([this](std::vector<byte>& fileData) {
		m_pathTransformShader = fileData;
	});

	auto createTestComputeCSTask = DX::ReadDataAsync(L"TestCompute.cso").then([this](std::vector<byte>& fileData) {
		m_testComputeShader = fileData;
	});

	// Create the pipeline state once the shaders are loaded.
	auto createPipelineStateTask = (createPSTask && createVSTask).then([this]() {

		static const D3D12_INPUT_ELEMENT_DESC inputLayout[] =
		{
			{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D12_INPUT_CLASSIFICATION_PER_VERTEX_DATA, 0 },
			{ "COLOR", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D12_INPUT_CLASSIFICATION_PER_VERTEX_DATA, 0 },
		};

		D3D12_GRAPHICS_PIPELINE_STATE_DESC state = {};
		state.InputLayout = { inputLayout, _countof(inputLayout) };
		state.pRootSignature = m_rootSignature.Get();
        state.VS = CD3DX12_SHADER_BYTECODE(&m_vertexShader[0], m_vertexShader.size());
        state.PS = CD3DX12_SHADER_BYTECODE(&m_pixelShader[0], m_pixelShader.size());
		state.RasterizerState = CD3DX12_RASTERIZER_DESC(D3D12_DEFAULT);
		state.BlendState = CD3DX12_BLEND_DESC(D3D12_DEFAULT);
		state.DepthStencilState = CD3DX12_DEPTH_STENCIL_DESC(D3D12_DEFAULT);
		state.SampleMask = UINT_MAX;
		state.PrimitiveTopologyType = D3D12_PRIMITIVE_TOPOLOGY_TYPE_TRIANGLE;
		state.NumRenderTargets = 1;
		state.RTVFormats[0] = m_deviceResources->GetBackBufferFormat();
		state.DSVFormat = m_deviceResources->GetDepthBufferFormat();
		state.SampleDesc.Count = 1;

		DX::ThrowIfFailed(m_deviceResources->GetD3DDevice()->CreateGraphicsPipelineState(&state, IID_PPV_ARGS(&m_pipelineState)));

		// Shader data can be deleted once the pipeline state is created.
		m_vertexShader.clear();
		m_pixelShader.clear();
	});

	auto createPathTransformAssets = createPathTransformCSTask.then([this]()
	{
		auto d3dDevice = m_deviceResources->GetD3DDevice();

		constexpr size_t numPaths = 1;
		constexpr size_t numVisibilityInts = numPaths * 8;
		constexpr size_t visibilitySizeInBytes = sizeof(int) * numVisibilityInts;

		D3D12_HEAP_DESC heapDesc;

		heapDesc.SizeInBytes = 1024 * 1024 * 64;
		heapDesc.Properties.Type = D3D12_HEAP_TYPE_DEFAULT;
		heapDesc.Properties.CPUPageProperty = D3D12_CPU_PAGE_PROPERTY_UNKNOWN;
		heapDesc.Properties.MemoryPoolPreference = D3D12_MEMORY_POOL_UNKNOWN;
		heapDesc.Properties.CreationNodeMask = 0;
		heapDesc.Properties.VisibleNodeMask = 0;
		heapDesc.Alignment = D3D12_DEFAULT_MSAA_RESOURCE_PLACEMENT_ALIGNMENT;
		heapDesc.Flags = D3D12_HEAP_FLAG_ALLOW_ONLY_BUFFERS;

		DX::ThrowIfFailed(d3dDevice->CreateHeap(&heapDesc, IID_PPV_ARGS(&m_heapDeviceLocal)));
		NAME_D3D12_OBJECT(m_heapDeviceLocal);

		heapDesc.Properties.Type = D3D12_HEAP_TYPE_UPLOAD;

		DX::ThrowIfFailed(d3dDevice->CreateHeap(&heapDesc, IID_PPV_ARGS(&m_heapUpload)));
		NAME_D3D12_OBJECT(m_heapUpload);

		D3D12_COMMAND_QUEUE_DESC queueDesc;

		queueDesc.Type = D3D12_COMMAND_LIST_TYPE_COMPUTE;
		queueDesc.Priority = D3D12_COMMAND_QUEUE_PRIORITY_NORMAL;
		queueDesc.Flags = D3D12_COMMAND_QUEUE_FLAG_NONE;
		queueDesc.NodeMask = 0;

		DX::ThrowIfFailed(d3dDevice->CreateCommandQueue(&queueDesc, IID_PPV_ARGS(&m_computeQueue)));
		NAME_D3D12_OBJECT(m_computeQueue);

		queueDesc.Type = D3D12_COMMAND_LIST_TYPE_COPY;

		DX::ThrowIfFailed(d3dDevice->CreateCommandQueue(&queueDesc, IID_PPV_ARGS(&m_copyQueue)));
		NAME_D3D12_OBJECT(m_copyQueue);

		D3D12_RESOURCE_DESC rsrcDesc;

		rsrcDesc.Dimension = D3D12_RESOURCE_DIMENSION_BUFFER;
		rsrcDesc.Alignment = D3D12_DEFAULT_RESOURCE_PLACEMENT_ALIGNMENT;
		rsrcDesc.Width = sizeof(g_pathVertices);
		rsrcDesc.Height = 1;
		rsrcDesc.DepthOrArraySize = 1;
		rsrcDesc.MipLevels = 1;
		rsrcDesc.Format = DXGI_FORMAT_UNKNOWN;
		rsrcDesc.SampleDesc.Count = 1;
		rsrcDesc.SampleDesc.Quality = 0;
		rsrcDesc.Layout = D3D12_TEXTURE_LAYOUT_ROW_MAJOR;
		rsrcDesc.Flags = D3D12_RESOURCE_FLAG_ALLOW_UNORDERED_ACCESS;

		D3D12_RESOURCE_ALLOCATION_INFO allocInfo = d3dDevice->GetResourceAllocationInfo(0, 1, &rsrcDesc);

		rsrcDesc.Alignment = allocInfo.Alignment;
		rsrcDesc.Width = allocInfo.SizeInBytes;

		DX::ThrowIfFailed(d3dDevice->CreatePlacedResource(m_heapDeviceLocal.Get(), 0, &rsrcDesc, D3D12_RESOURCE_STATE_COMMON, nullptr, IID_PPV_ARGS(&m_pathVertexBuffer)));
		NAME_D3D12_OBJECT(m_pathVertexBuffer);

		rsrcDesc.Flags = D3D12_RESOURCE_FLAG_NONE;

		DX::ThrowIfFailed(d3dDevice->CreatePlacedResource(m_heapUpload.Get(), 0, &rsrcDesc, D3D12_RESOURCE_STATE_GENERIC_READ, nullptr, IID_PPV_ARGS(&m_pathVertexBufferStaging)));
		NAME_D3D12_OBJECT(m_pathVertexBufferStaging);

		D3D12_RANGE mapRange;

		mapRange.Begin = 0;
		mapRange.End = sizeof(g_pathVertices);

		void *mappedData = nullptr;

		DX::ThrowIfFailed(m_pathVertexBufferStaging->Map(0, &mapRange, &mappedData));

		memcpy_s(mappedData, sizeof(g_pathVertices), g_pathVertices, sizeof(g_pathVertices));

		m_pathVertexBufferStaging->Unmap(0, &mapRange);

		UINT64 offset = rsrcDesc.Width;

		rsrcDesc.Width = sizeof(SVec2DF) * _countof(g_pathVertices);
		rsrcDesc.Flags = D3D12_RESOURCE_FLAG_ALLOW_UNORDERED_ACCESS;

		allocInfo = d3dDevice->GetResourceAllocationInfo(0, 1, &rsrcDesc);

		rsrcDesc.Alignment = allocInfo.Alignment;
		rsrcDesc.Width = allocInfo.SizeInBytes;

		offset = Align(offset, rsrcDesc.Alignment);

		DX::ThrowIfFailed(d3dDevice->CreatePlacedResource(m_heapDeviceLocal.Get(), offset, &rsrcDesc, D3D12_RESOURCE_STATE_COMMON, nullptr, IID_PPV_ARGS(&m_pathTransformedVertices)));
		NAME_D3D12_OBJECT(m_pathTransformedVertices);

		offset += rsrcDesc.Width;

		rsrcDesc.Width = visibilitySizeInBytes;

		allocInfo = d3dDevice->GetResourceAllocationInfo(0, 1, &rsrcDesc);

		rsrcDesc.Alignment = allocInfo.Alignment;
		rsrcDesc.Width = allocInfo.SizeInBytes;

		offset = Align(offset, rsrcDesc.Alignment);

		DX::ThrowIfFailed(d3dDevice->CreatePlacedResource(m_heapDeviceLocal.Get(), offset, &rsrcDesc, D3D12_RESOURCE_STATE_COMMON, nullptr, IID_PPV_ARGS(&m_pathVisibility)));
		NAME_D3D12_OBJECT(m_pathVisibility);

		D3D12_DESCRIPTOR_HEAP_DESC descriptorHeapDesc;

		descriptorHeapDesc.Type = D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV;
		descriptorHeapDesc.NumDescriptors = 256;
		descriptorHeapDesc.Flags = D3D12_DESCRIPTOR_HEAP_FLAG_SHADER_VISIBLE;
		descriptorHeapDesc.NodeMask = 0;

		DX::ThrowIfFailed(d3dDevice->CreateDescriptorHeap(&descriptorHeapDesc, IID_PPV_ARGS(&m_pathCBVHeap)));
		NAME_D3D12_OBJECT(m_pathCBVHeap);

		D3D12_DESCRIPTOR_RANGE1 descriptorRanges[3];

		descriptorRanges[0].RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_SRV;
		descriptorRanges[0].NumDescriptors = 1;
		descriptorRanges[0].BaseShaderRegister = 0;
		descriptorRanges[0].RegisterSpace = 1;
		descriptorRanges[0].Flags = D3D12_DESCRIPTOR_RANGE_FLAG_NONE;
		descriptorRanges[0].OffsetInDescriptorsFromTableStart = D3D12_DESCRIPTOR_RANGE_OFFSET_APPEND;

		descriptorRanges[1].RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_UAV;
		descriptorRanges[1].NumDescriptors = 1;
		descriptorRanges[1].BaseShaderRegister = 0;
		descriptorRanges[1].RegisterSpace = 1;
		descriptorRanges[1].Flags = D3D12_DESCRIPTOR_RANGE_FLAG_NONE;
		descriptorRanges[1].OffsetInDescriptorsFromTableStart = D3D12_DESCRIPTOR_RANGE_OFFSET_APPEND;

		descriptorRanges[2].RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_UAV;
		descriptorRanges[2].NumDescriptors = 1;
		descriptorRanges[2].BaseShaderRegister = 1;
		descriptorRanges[2].RegisterSpace = 1;
		descriptorRanges[2].Flags = D3D12_DESCRIPTOR_RANGE_FLAG_NONE;
		descriptorRanges[2].OffsetInDescriptorsFromTableStart = D3D12_DESCRIPTOR_RANGE_OFFSET_APPEND;

		D3D12_ROOT_PARAMETER1 rootParams[2];

		rootParams[0].ParameterType = D3D12_ROOT_PARAMETER_TYPE_32BIT_CONSTANTS;
		rootParams[0].Constants.ShaderRegister = 0;
		rootParams[0].Constants.RegisterSpace = 0;
		rootParams[0].Constants.Num32BitValues = (sizeof(XMFLOAT4X4)  + sizeof(XMFLOAT4) * 2) / sizeof(int);
		rootParams[0].ShaderVisibility = D3D12_SHADER_VISIBILITY_ALL;

		rootParams[1].ParameterType = D3D12_ROOT_PARAMETER_TYPE_DESCRIPTOR_TABLE;
		rootParams[1].DescriptorTable.NumDescriptorRanges = 3;
		rootParams[1].DescriptorTable.pDescriptorRanges = descriptorRanges;
		rootParams[1].ShaderVisibility = D3D12_SHADER_VISIBILITY_ALL;

		D3D12_VERSIONED_ROOT_SIGNATURE_DESC rootSigDesc;

		rootSigDesc.Version = D3D_ROOT_SIGNATURE_VERSION_1_1;
		rootSigDesc.Desc_1_1.NumParameters = 2;
		rootSigDesc.Desc_1_1.pParameters = rootParams;
		rootSigDesc.Desc_1_1.NumStaticSamplers = 0;
		rootSigDesc.Desc_1_1.pStaticSamplers = nullptr;
		rootSigDesc.Desc_1_1.Flags =
			D3D12_ROOT_SIGNATURE_FLAG_DENY_VERTEX_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_HULL_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_DOMAIN_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_GEOMETRY_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_PIXEL_SHADER_ROOT_ACCESS;

		Microsoft::WRL::ComPtr<ID3DBlob> rootSigBlob;
		Microsoft::WRL::ComPtr<ID3DBlob> rootSigErrorBlob;

		DX::ThrowIfFailed(D3D12SerializeVersionedRootSignature(&rootSigDesc, &rootSigBlob, &rootSigErrorBlob));

		DX::ThrowIfFailed(d3dDevice->CreateRootSignature(0, rootSigBlob->GetBufferPointer(), rootSigBlob->GetBufferSize(), IID_PPV_ARGS(&m_pathTransformVerticesRootSignature)));
		NAME_D3D12_OBJECT(m_pathTransformVerticesRootSignature);

		D3D12_COMPUTE_PIPELINE_STATE_DESC computePipelineDesc;

		computePipelineDesc.pRootSignature = m_pathTransformVerticesRootSignature.Get();
		computePipelineDesc.CS.pShaderBytecode = m_pathTransformShader.data();
		computePipelineDesc.CS.BytecodeLength = m_pathTransformShader.size();
		computePipelineDesc.CachedPSO.pCachedBlob = nullptr;
		computePipelineDesc.CachedPSO.CachedBlobSizeInBytes = 0;
		computePipelineDesc.Flags = D3D12_PIPELINE_STATE_FLAG_NONE;
		computePipelineDesc.NodeMask = 0;

		DX::ThrowIfFailed(d3dDevice->CreateComputePipelineState(&computePipelineDesc, IID_PPV_ARGS(&m_pathTransformVerticesPipeline)));
		NAME_D3D12_OBJECT(m_pathTransformVerticesPipeline);

		D3D12_SHADER_RESOURCE_VIEW_DESC srvDesc;

		srvDesc.Format = DXGI_FORMAT_R32G32B32_FLOAT;
		srvDesc.ViewDimension = D3D12_SRV_DIMENSION_BUFFER;
		srvDesc.Shader4ComponentMapping = D3D12_ENCODE_SHADER_4_COMPONENT_MAPPING(D3D12_SHADER_COMPONENT_MAPPING_FROM_MEMORY_COMPONENT_0, D3D12_SHADER_COMPONENT_MAPPING_FROM_MEMORY_COMPONENT_1, D3D12_SHADER_COMPONENT_MAPPING_FROM_MEMORY_COMPONENT_2, D3D12_SHADER_COMPONENT_MAPPING_FROM_MEMORY_COMPONENT_3);
		srvDesc.Buffer.FirstElement = 0;
		srvDesc.Buffer.NumElements = _countof(g_pathVertices);
		srvDesc.Buffer.StructureByteStride = 0;
		srvDesc.Buffer.Flags = D3D12_BUFFER_SRV_FLAG_NONE;

		const D3D12_CPU_DESCRIPTOR_HANDLE cpuStart = m_pathCBVHeap->GetCPUDescriptorHandleForHeapStart();
		const UINT descriptorSize = d3dDevice->GetDescriptorHandleIncrementSize(D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV);

		D3D12_CPU_DESCRIPTOR_HANDLE descriptorHandle = cpuStart;

		d3dDevice->CreateShaderResourceView(m_pathVertexBuffer.Get(), &srvDesc, descriptorHandle);

		descriptorHandle.ptr += descriptorSize;

		D3D12_UNORDERED_ACCESS_VIEW_DESC uavDesc;

		uavDesc.Format = DXGI_FORMAT_R32G32_FLOAT;
		uavDesc.ViewDimension = D3D12_UAV_DIMENSION_BUFFER;
		uavDesc.Buffer.FirstElement = 0;
		uavDesc.Buffer.NumElements = _countof(g_pathVertices);
		uavDesc.Buffer.StructureByteStride = 0;
		uavDesc.Buffer.CounterOffsetInBytes = 0;
		uavDesc.Buffer.Flags = D3D12_BUFFER_UAV_FLAG_NONE;

		d3dDevice->CreateUnorderedAccessView(m_pathTransformedVertices.Get(), nullptr, &uavDesc, descriptorHandle);

		descriptorHandle.ptr += descriptorSize;

		uavDesc.Format = DXGI_FORMAT_R32_UINT;
		uavDesc.Buffer.NumElements = numVisibilityInts;

		d3dDevice->CreateUnorderedAccessView(m_pathVisibility.Get(), nullptr, &uavDesc, descriptorHandle);

		DX::ThrowIfFailed(d3dDevice->CreateCommandAllocator(D3D12_COMMAND_LIST_TYPE_COMPUTE, IID_PPV_ARGS(&m_pathTransformComputeListAllocator)));
		NAME_D3D12_OBJECT(m_pathTransformComputeListAllocator);

		DX::ThrowIfFailed(d3dDevice->CreateCommandList(0, D3D12_COMMAND_LIST_TYPE_COMPUTE, m_pathTransformComputeListAllocator.Get(), m_pathTransformVerticesPipeline.Get(), IID_PPV_ARGS(&m_pathTransformComputeList)));
		NAME_D3D12_OBJECT(m_pathTransformComputeList);

		DX::ThrowIfFailed(d3dDevice->CreateFence(0, D3D12_FENCE_FLAG_NONE, IID_PPV_ARGS(&m_pathTransformFence)));
		NAME_D3D12_OBJECT(m_pathTransformFence);
	});

	auto createTestComputeAssets = createTestComputeCSTask.then([this]()
	{
		auto d3dDevice = m_deviceResources->GetD3DDevice();

		D3D12_RESOURCE_DESC rsrcDesc;

		rsrcDesc.Dimension = D3D12_RESOURCE_DIMENSION_BUFFER;
		rsrcDesc.Alignment = D3D12_DEFAULT_RESOURCE_PLACEMENT_ALIGNMENT;
		rsrcDesc.Width = sizeof(int) * TestComputeNumBufferItems;
		rsrcDesc.Height = 1;
		rsrcDesc.DepthOrArraySize = 1;
		rsrcDesc.MipLevels = 1;
		rsrcDesc.Format = DXGI_FORMAT_UNKNOWN;
		rsrcDesc.SampleDesc.Count = 1;
		rsrcDesc.SampleDesc.Quality = 0;
		rsrcDesc.Layout = D3D12_TEXTURE_LAYOUT_ROW_MAJOR;
		rsrcDesc.Flags = D3D12_RESOURCE_FLAG_ALLOW_UNORDERED_ACCESS;

		D3D12_RESOURCE_ALLOCATION_INFO allocInfo = d3dDevice->GetResourceAllocationInfo(0, 1, &rsrcDesc);

		rsrcDesc.Alignment = allocInfo.Alignment;
		rsrcDesc.Width = allocInfo.SizeInBytes;

		D3D12_HEAP_PROPERTIES heapProperties;

		heapProperties.Type = D3D12_HEAP_TYPE_DEFAULT;
		heapProperties.CPUPageProperty = D3D12_CPU_PAGE_PROPERTY_UNKNOWN;
		heapProperties.MemoryPoolPreference = D3D12_MEMORY_POOL_UNKNOWN;
		heapProperties.CreationNodeMask = 0;
		heapProperties.VisibleNodeMask = 0;

		DX::ThrowIfFailed(d3dDevice->CreateCommittedResource(&heapProperties, D3D12_HEAP_FLAG_NONE, &rsrcDesc, D3D12_RESOURCE_STATE_COMMON, nullptr, IID_PPV_ARGS(&m_testComputeOutput)));
		NAME_D3D12_OBJECT(m_testComputeOutput);

		heapProperties.Type = D3D12_HEAP_TYPE_READBACK;
		rsrcDesc.Flags = D3D12_RESOURCE_FLAG_NONE;

		DX::ThrowIfFailed(d3dDevice->CreateCommittedResource(&heapProperties, D3D12_HEAP_FLAG_NONE, &rsrcDesc, D3D12_RESOURCE_STATE_COPY_DEST, nullptr, IID_PPV_ARGS(&m_testComputeOutputDownload)));
		NAME_D3D12_OBJECT(m_testComputeOutputDownload);

		D3D12_DESCRIPTOR_HEAP_DESC descriptorHeapDesc;

		descriptorHeapDesc.Type = D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV;
		descriptorHeapDesc.NumDescriptors = 1;
		descriptorHeapDesc.Flags = D3D12_DESCRIPTOR_HEAP_FLAG_SHADER_VISIBLE;
		descriptorHeapDesc.NodeMask = 0;

		DX::ThrowIfFailed(d3dDevice->CreateDescriptorHeap(&descriptorHeapDesc, IID_PPV_ARGS(&m_testComputeCBVHeap)));
		NAME_D3D12_OBJECT(m_testComputeCBVHeap);

		D3D12_DESCRIPTOR_RANGE1 descriptorRanges[3];

		descriptorRanges[0].RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_UAV;
		descriptorRanges[0].NumDescriptors = 1;
		descriptorRanges[0].BaseShaderRegister = 0;
		descriptorRanges[0].RegisterSpace = 1;
		descriptorRanges[0].Flags = D3D12_DESCRIPTOR_RANGE_FLAG_NONE;
		descriptorRanges[0].OffsetInDescriptorsFromTableStart = D3D12_DESCRIPTOR_RANGE_OFFSET_APPEND;

		/*descriptorRanges[1].RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_UAV;
		descriptorRanges[1].NumDescriptors = 1;
		descriptorRanges[1].BaseShaderRegister = 0;
		descriptorRanges[1].RegisterSpace = 1;
		descriptorRanges[1].Flags = D3D12_DESCRIPTOR_RANGE_FLAG_NONE;
		descriptorRanges[1].OffsetInDescriptorsFromTableStart = D3D12_DESCRIPTOR_RANGE_OFFSET_APPEND;

		descriptorRanges[2].RangeType = D3D12_DESCRIPTOR_RANGE_TYPE_UAV;
		descriptorRanges[2].NumDescriptors = 1;
		descriptorRanges[2].BaseShaderRegister = 1;
		descriptorRanges[2].RegisterSpace = 1;
		descriptorRanges[2].Flags = D3D12_DESCRIPTOR_RANGE_FLAG_NONE;
		descriptorRanges[2].OffsetInDescriptorsFromTableStart = D3D12_DESCRIPTOR_RANGE_OFFSET_APPEND;*/

		D3D12_ROOT_PARAMETER1 rootParams[2];

		rootParams[0].ParameterType = D3D12_ROOT_PARAMETER_TYPE_32BIT_CONSTANTS;
		rootParams[0].Constants.ShaderRegister = 0;
		rootParams[0].Constants.RegisterSpace = 0;
		rootParams[0].Constants.Num32BitValues = 1;
		rootParams[0].ShaderVisibility = D3D12_SHADER_VISIBILITY_ALL;

		rootParams[1].ParameterType = D3D12_ROOT_PARAMETER_TYPE_DESCRIPTOR_TABLE;
		rootParams[1].DescriptorTable.NumDescriptorRanges = 1;
		rootParams[1].DescriptorTable.pDescriptorRanges = descriptorRanges;
		rootParams[1].ShaderVisibility = D3D12_SHADER_VISIBILITY_ALL;

		D3D12_VERSIONED_ROOT_SIGNATURE_DESC rootSigDesc;

		rootSigDesc.Version = D3D_ROOT_SIGNATURE_VERSION_1_1;
		rootSigDesc.Desc_1_1.NumParameters = 2;
		rootSigDesc.Desc_1_1.pParameters = rootParams;
		rootSigDesc.Desc_1_1.NumStaticSamplers = 0;
		rootSigDesc.Desc_1_1.pStaticSamplers = nullptr;
		rootSigDesc.Desc_1_1.Flags =
			D3D12_ROOT_SIGNATURE_FLAG_DENY_VERTEX_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_HULL_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_DOMAIN_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_GEOMETRY_SHADER_ROOT_ACCESS
			| D3D12_ROOT_SIGNATURE_FLAG_DENY_PIXEL_SHADER_ROOT_ACCESS;

		Microsoft::WRL::ComPtr<ID3DBlob> rootSigBlob;
		Microsoft::WRL::ComPtr<ID3DBlob> rootSigErrorBlob;

		DX::ThrowIfFailed(D3D12SerializeVersionedRootSignature(&rootSigDesc, &rootSigBlob, &rootSigErrorBlob));

		DX::ThrowIfFailed(d3dDevice->CreateRootSignature(0, rootSigBlob->GetBufferPointer(), rootSigBlob->GetBufferSize(), IID_PPV_ARGS(&m_testComputeRootSignature)));
		NAME_D3D12_OBJECT(m_testComputeRootSignature);

		D3D12_COMPUTE_PIPELINE_STATE_DESC computePipelineDesc;

		computePipelineDesc.pRootSignature = m_testComputeRootSignature.Get();
		computePipelineDesc.CS.pShaderBytecode = m_testComputeShader.data();
		computePipelineDesc.CS.BytecodeLength = m_testComputeShader.size();
		computePipelineDesc.CachedPSO.pCachedBlob = nullptr;
		computePipelineDesc.CachedPSO.CachedBlobSizeInBytes = 0;
		computePipelineDesc.Flags = D3D12_PIPELINE_STATE_FLAG_NONE;
		computePipelineDesc.NodeMask = 0;

		DX::ThrowIfFailed(d3dDevice->CreateComputePipelineState(&computePipelineDesc, IID_PPV_ARGS(&m_testComputePipeline)));
		NAME_D3D12_OBJECT(m_testComputePipeline);

		const D3D12_CPU_DESCRIPTOR_HANDLE cpuStart = m_testComputeCBVHeap->GetCPUDescriptorHandleForHeapStart();
		const UINT descriptorSize = d3dDevice->GetDescriptorHandleIncrementSize(D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV);

		D3D12_CPU_DESCRIPTOR_HANDLE descriptorHandle = cpuStart;

		D3D12_UNORDERED_ACCESS_VIEW_DESC uavDesc;

		//uavDesc.Format = DXGI_FORMAT_R32_UINT;
		uavDesc.Format = DXGI_FORMAT_UNKNOWN;
		uavDesc.ViewDimension = D3D12_UAV_DIMENSION_BUFFER;
		uavDesc.Buffer.FirstElement = 0;
		uavDesc.Buffer.NumElements = TestComputeNumBufferItems;
		uavDesc.Buffer.StructureByteStride = sizeof(int);
		uavDesc.Buffer.CounterOffsetInBytes = 0;
		uavDesc.Buffer.Flags = D3D12_BUFFER_UAV_FLAG_NONE;

		d3dDevice->CreateUnorderedAccessView(m_testComputeOutput.Get(), nullptr, &uavDesc, descriptorHandle);

		DX::ThrowIfFailed(d3dDevice->CreateCommandAllocator(D3D12_COMMAND_LIST_TYPE_COMPUTE, IID_PPV_ARGS(&m_testComputeListAllocator)));
		NAME_D3D12_OBJECT(m_testComputeListAllocator);

		DX::ThrowIfFailed(d3dDevice->CreateCommandList(0, D3D12_COMMAND_LIST_TYPE_COMPUTE, m_testComputeListAllocator.Get(), m_testComputePipeline.Get(), IID_PPV_ARGS(&m_testComputeList)));
		NAME_D3D12_OBJECT(m_testComputeList);

		DX::ThrowIfFailed(d3dDevice->CreateFence(0, D3D12_FENCE_FLAG_NONE, IID_PPV_ARGS(&m_testComputeFence)));
		NAME_D3D12_OBJECT(m_testComputeFence);
	});

	// Create and upload cube geometry resources to the GPU.
	auto createAssetsTask = createPipelineStateTask.then([this]() {
		auto d3dDevice = m_deviceResources->GetD3DDevice();

		// Create a command list.
		DX::ThrowIfFailed(d3dDevice->CreateCommandList(0, D3D12_COMMAND_LIST_TYPE_DIRECT, m_deviceResources->GetCommandAllocator(), m_pipelineState.Get(), IID_PPV_ARGS(&m_commandList)));
        NAME_D3D12_OBJECT(m_commandList);

		// Cube vertices. Each vertex has a position and a color.
		VertexPositionColor cubeVertices[] =
		{
			{ XMFLOAT3(-0.5f, -0.5f, -0.5f), XMFLOAT3(0.0f, 0.0f, 0.0f) },
			{ XMFLOAT3(-0.5f, -0.5f,  0.5f), XMFLOAT3(0.0f, 0.0f, 1.0f) },
			{ XMFLOAT3(-0.5f,  0.5f, -0.5f), XMFLOAT3(0.0f, 1.0f, 0.0f) },
			{ XMFLOAT3(-0.5f,  0.5f,  0.5f), XMFLOAT3(0.0f, 1.0f, 1.0f) },
			{ XMFLOAT3(0.5f, -0.5f, -0.5f), XMFLOAT3(1.0f, 0.0f, 0.0f) },
			{ XMFLOAT3(0.5f, -0.5f,  0.5f), XMFLOAT3(1.0f, 0.0f, 1.0f) },
			{ XMFLOAT3(0.5f,  0.5f, -0.5f), XMFLOAT3(1.0f, 1.0f, 0.0f) },
			{ XMFLOAT3(0.5f,  0.5f,  0.5f), XMFLOAT3(1.0f, 1.0f, 1.0f) },
		};

		const UINT vertexBufferSize = sizeof(cubeVertices);

		// Create the vertex buffer resource in the GPU's default heap and copy vertex data into it using the upload heap.
		// The upload resource must not be released until after the GPU has finished using it.
		Microsoft::WRL::ComPtr<ID3D12Resource> vertexBufferUpload;

		CD3DX12_HEAP_PROPERTIES defaultHeapProperties(D3D12_HEAP_TYPE_DEFAULT);
		CD3DX12_RESOURCE_DESC vertexBufferDesc = CD3DX12_RESOURCE_DESC::Buffer(vertexBufferSize);
		DX::ThrowIfFailed(d3dDevice->CreateCommittedResource(
			&defaultHeapProperties,
			D3D12_HEAP_FLAG_NONE,
			&vertexBufferDesc,
			D3D12_RESOURCE_STATE_COPY_DEST,
			nullptr,
			IID_PPV_ARGS(&m_vertexBuffer)));

		CD3DX12_HEAP_PROPERTIES uploadHeapProperties(D3D12_HEAP_TYPE_UPLOAD);
		DX::ThrowIfFailed(d3dDevice->CreateCommittedResource(
			&uploadHeapProperties,
			D3D12_HEAP_FLAG_NONE,
			&vertexBufferDesc,
			D3D12_RESOURCE_STATE_GENERIC_READ,
			nullptr,
			IID_PPV_ARGS(&vertexBufferUpload)));

        NAME_D3D12_OBJECT(m_vertexBuffer);

		// Upload the vertex buffer to the GPU.
		{
			D3D12_SUBRESOURCE_DATA vertexData = {};
			vertexData.pData = reinterpret_cast<BYTE*>(cubeVertices);
			vertexData.RowPitch = vertexBufferSize;
			vertexData.SlicePitch = vertexData.RowPitch;

			UpdateSubresources(m_commandList.Get(), m_vertexBuffer.Get(), vertexBufferUpload.Get(), 0, 0, 1, &vertexData);

			CD3DX12_RESOURCE_BARRIER vertexBufferResourceBarrier =
				CD3DX12_RESOURCE_BARRIER::Transition(m_vertexBuffer.Get(), D3D12_RESOURCE_STATE_COPY_DEST, D3D12_RESOURCE_STATE_VERTEX_AND_CONSTANT_BUFFER);
			m_commandList->ResourceBarrier(1, &vertexBufferResourceBarrier);
		}

		// Load mesh indices. Each trio of indices represents a triangle to be rendered on the screen.
		// For example: 0,2,1 means that the vertices with indexes 0, 2 and 1 from the vertex buffer compose the
		// first triangle of this mesh.
		unsigned short cubeIndices[] =
		{
			0, 2, 1, // -x
			1, 2, 3,

			4, 5, 6, // +x
			5, 7, 6,

			0, 1, 5, // -y
			0, 5, 4,

			2, 6, 7, // +y
			2, 7, 3,

			0, 4, 6, // -z
			0, 6, 2,

			1, 3, 7, // +z
			1, 7, 5,
		};

		const UINT indexBufferSize = sizeof(cubeIndices);

		// Create the index buffer resource in the GPU's default heap and copy index data into it using the upload heap.
		// The upload resource must not be released until after the GPU has finished using it.
		Microsoft::WRL::ComPtr<ID3D12Resource> indexBufferUpload;

		CD3DX12_RESOURCE_DESC indexBufferDesc = CD3DX12_RESOURCE_DESC::Buffer(indexBufferSize);
		DX::ThrowIfFailed(d3dDevice->CreateCommittedResource(
			&defaultHeapProperties,
			D3D12_HEAP_FLAG_NONE,
			&indexBufferDesc,
			D3D12_RESOURCE_STATE_COPY_DEST,
			nullptr,
			IID_PPV_ARGS(&m_indexBuffer)));

		DX::ThrowIfFailed(d3dDevice->CreateCommittedResource(
			&uploadHeapProperties,
			D3D12_HEAP_FLAG_NONE,
			&indexBufferDesc,
			D3D12_RESOURCE_STATE_GENERIC_READ,
			nullptr,
			IID_PPV_ARGS(&indexBufferUpload)));

		NAME_D3D12_OBJECT(m_indexBuffer);

		// Upload the index buffer to the GPU.
		{
			D3D12_SUBRESOURCE_DATA indexData = {};
			indexData.pData = reinterpret_cast<BYTE*>(cubeIndices);
			indexData.RowPitch = indexBufferSize;
			indexData.SlicePitch = indexData.RowPitch;

			UpdateSubresources(m_commandList.Get(), m_indexBuffer.Get(), indexBufferUpload.Get(), 0, 0, 1, &indexData);

			CD3DX12_RESOURCE_BARRIER indexBufferResourceBarrier =
				CD3DX12_RESOURCE_BARRIER::Transition(m_indexBuffer.Get(), D3D12_RESOURCE_STATE_COPY_DEST, D3D12_RESOURCE_STATE_INDEX_BUFFER);
			m_commandList->ResourceBarrier(1, &indexBufferResourceBarrier);
		}

		// Create a descriptor heap for the constant buffers.
		{
			D3D12_DESCRIPTOR_HEAP_DESC heapDesc = {};
			heapDesc.NumDescriptors = DX::c_frameCount;
			heapDesc.Type = D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV;
			// This flag indicates that this descriptor heap can be bound to the pipeline and that descriptors contained in it can be referenced by a root table.
			heapDesc.Flags = D3D12_DESCRIPTOR_HEAP_FLAG_SHADER_VISIBLE;
			DX::ThrowIfFailed(d3dDevice->CreateDescriptorHeap(&heapDesc, IID_PPV_ARGS(&m_cbvHeap)));

            NAME_D3D12_OBJECT(m_cbvHeap);
		}

		CD3DX12_RESOURCE_DESC constantBufferDesc = CD3DX12_RESOURCE_DESC::Buffer(DX::c_frameCount * c_alignedConstantBufferSize);
		DX::ThrowIfFailed(d3dDevice->CreateCommittedResource(
			&uploadHeapProperties,
			D3D12_HEAP_FLAG_NONE,
			&constantBufferDesc,
			D3D12_RESOURCE_STATE_GENERIC_READ,
			nullptr,
			IID_PPV_ARGS(&m_constantBuffer)));

        NAME_D3D12_OBJECT(m_constantBuffer);

		// Create constant buffer views to access the upload buffer.
		D3D12_GPU_VIRTUAL_ADDRESS cbvGpuAddress = m_constantBuffer->GetGPUVirtualAddress();
		CD3DX12_CPU_DESCRIPTOR_HANDLE cbvCpuHandle(m_cbvHeap->GetCPUDescriptorHandleForHeapStart());
		m_cbvDescriptorSize = d3dDevice->GetDescriptorHandleIncrementSize(D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV);

		for (int n = 0; n < DX::c_frameCount; n++)
		{
			D3D12_CONSTANT_BUFFER_VIEW_DESC desc;
			desc.BufferLocation = cbvGpuAddress;
			desc.SizeInBytes = c_alignedConstantBufferSize;
			d3dDevice->CreateConstantBufferView(&desc, cbvCpuHandle);

			cbvGpuAddress += desc.SizeInBytes;
			cbvCpuHandle.Offset(m_cbvDescriptorSize);
		}

		// Map the constant buffers.
		CD3DX12_RANGE readRange(0, 0);		// We do not intend to read from this resource on the CPU.
		DX::ThrowIfFailed(m_constantBuffer->Map(0, &readRange, reinterpret_cast<void**>(&m_mappedConstantBuffer)));
		ZeroMemory(m_mappedConstantBuffer, DX::c_frameCount * c_alignedConstantBufferSize);
		// We don't unmap this until the app closes. Keeping things mapped for the lifetime of the resource is okay.

		// Close the command list and execute it to begin the vertex/index buffer copy into the GPU's default heap.
		DX::ThrowIfFailed(m_commandList->Close());
		ID3D12CommandList* ppCommandLists[] = { m_commandList.Get() };
		m_deviceResources->GetCommandQueue()->ExecuteCommandLists(_countof(ppCommandLists), ppCommandLists);

		// Create vertex/index buffer views.
		m_vertexBufferView.BufferLocation = m_vertexBuffer->GetGPUVirtualAddress();
		m_vertexBufferView.StrideInBytes = sizeof(VertexPositionColor);
		m_vertexBufferView.SizeInBytes = sizeof(cubeVertices);

		m_indexBufferView.BufferLocation = m_indexBuffer->GetGPUVirtualAddress();
		m_indexBufferView.SizeInBytes = sizeof(cubeIndices);
		m_indexBufferView.Format = DXGI_FORMAT_R16_UINT;

		// Wait for the command list to finish executing; the vertex/index buffers need to be uploaded to the GPU before the upload resources go out of scope.
		m_deviceResources->WaitForGpu();
	});

	(createPathTransformAssets && createAssetsTask && createTestComputeAssets).then([this]() {
		m_loadingComplete = true;
	});
}

// Initializes view parameters when the window size changes.
void Sample3DSceneRenderer::CreateWindowSizeDependentResources()
{
	Size outputSize = m_deviceResources->GetOutputSize();
	float aspectRatio = outputSize.Width / outputSize.Height;
	float fovAngleY = 70.0f * XM_PI / 180.0f;

	D3D12_VIEWPORT viewport = m_deviceResources->GetScreenViewport();
	m_scissorRect = { 0, 0, static_cast<LONG>(viewport.Width), static_cast<LONG>(viewport.Height)};

	// This is a simple example of change that can be made when the app is in
	// portrait or snapped view.
	if (aspectRatio < 1.0f)
	{
		fovAngleY *= 2.0f;
	}

	// Note that the OrientationTransform3D matrix is post-multiplied here
	// in order to correctly orient the scene to match the display orientation.
	// This post-multiplication step is required for any draw calls that are
	// made to the swap chain render target. For draw calls to other targets,
	// this transform should not be applied.

	// This sample makes use of a right-handed coordinate system using row-major matrices.
	XMMATRIX perspectiveMatrix = XMMatrixPerspectiveFovRH(
		fovAngleY,
		aspectRatio,
		0.01f,
		100.0f
		);

	XMFLOAT4X4 orientation = m_deviceResources->GetOrientationTransform3D();
	XMMATRIX orientationMatrix = XMLoadFloat4x4(&orientation);

	XMStoreFloat4x4(
		&m_constantBufferData.projection,
		XMMatrixTranspose(perspectiveMatrix * orientationMatrix)
		);

	// Eye is at (0,0.7,1.5), looking at point (0,-0.1,0) with the up-vector along the y-axis.
	static const XMVECTORF32 eye = { 0.0f, 0.7f, 1.5f, 0.0f };
	static const XMVECTORF32 at = { 0.0f, -0.1f, 0.0f, 0.0f };
	static const XMVECTORF32 up = { 0.0f, 1.0f, 0.0f, 0.0f };

	XMStoreFloat4x4(&m_constantBufferData.view, XMMatrixTranspose(XMMatrixLookAtRH(eye, at, up)));
}

// Called once per frame, rotates the cube and calculates the model and view matrices.
void Sample3DSceneRenderer::Update(DX::StepTimer const& timer)
{
	if (m_loadingComplete)
	{
		if (!m_tracking)
		{
			// Rotate the cube a small amount.
			m_angle += static_cast<float>(timer.GetElapsedSeconds()) * m_radiansPerSecond;

			Rotate(m_angle);
		}

		// Update the constant buffer resource.
		UINT8* destination = m_mappedConstantBuffer + (m_deviceResources->GetCurrentFrameIndex() * c_alignedConstantBufferSize);
		memcpy(destination, &m_constantBufferData, sizeof(m_constantBufferData));
	}
}

// Saves the current state of the renderer.
void Sample3DSceneRenderer::SaveState()
{
	auto state = ApplicationData::Current->LocalSettings->Values;

	if (state->HasKey(AngleKey))
	{
		state->Remove(AngleKey);
	}
	if (state->HasKey(TrackingKey))
	{
		state->Remove(TrackingKey);
	}

	state->Insert(AngleKey, PropertyValue::CreateSingle(m_angle));
	state->Insert(TrackingKey, PropertyValue::CreateBoolean(m_tracking));
}

// Restores the previous state of the renderer.
void Sample3DSceneRenderer::LoadState()
{
	auto state = ApplicationData::Current->LocalSettings->Values;
	if (state->HasKey(AngleKey))
	{
		m_angle = safe_cast<IPropertyValue^>(state->Lookup(AngleKey))->GetSingle();
		state->Remove(AngleKey);
	}
	if (state->HasKey(TrackingKey))
	{
		m_tracking = safe_cast<IPropertyValue^>(state->Lookup(TrackingKey))->GetBoolean();
		state->Remove(TrackingKey);
	}
}

// Rotate the 3D cube model a set amount of radians.
void Sample3DSceneRenderer::Rotate(float radians)
{
	// Prepare to pass the updated model matrix to the shader.
	XMStoreFloat4x4(&m_constantBufferData.model, XMMatrixTranspose(XMMatrixRotationY(radians)));
}

void Sample3DSceneRenderer::StartTracking()
{
	m_tracking = true;
}

// When tracking, the 3D cube can be rotated around its Y axis by tracking pointer position relative to the output screen width.
void Sample3DSceneRenderer::TrackingUpdate(float positionX)
{
	if (m_tracking)
	{
		float radians = XM_2PI * 2.0f * positionX / m_deviceResources->GetOutputSize().Width;
		Rotate(radians);
	}
}

void Sample3DSceneRenderer::StopTracking()
{
	m_tracking = false;
}

void Sample3DSceneRenderer::TransformVertices()
{
	if(m_transformedVertices)
	{
		return;
	}

	m_transformedVertices = true;

	PIXBeginEvent(m_pathTransformComputeList.Get(), 0, L"Transform Vertices");

	//m_pathTransformComputeList->SetPipelineState(m_pathTransformVerticesPipeline.Get());
	m_pathTransformComputeList->SetComputeRootSignature(m_pathTransformVerticesRootSignature.Get());

	CD3DX12_RESOURCE_BARRIER barriers[3];

	//barriers[0] = CD3DX12_RESOURCE_BARRIER::Transition(m_pathVertexBufferStaging.Get(), D3D12_RESOURCE_STATE_COMMON, D3D12_RESOURCE_STATE_COPY_SOURCE);
	barriers[0] = CD3DX12_RESOURCE_BARRIER::Transition(m_pathVertexBuffer.Get(), D3D12_RESOURCE_STATE_COMMON, D3D12_RESOURCE_STATE_COPY_DEST);

	m_pathTransformComputeList->ResourceBarrier(1, barriers);

	m_pathTransformComputeList->CopyResource(m_pathVertexBuffer.Get(), m_pathVertexBufferStaging.Get());

	barriers[0] = CD3DX12_RESOURCE_BARRIER::Transition(m_pathVertexBuffer.Get(), D3D12_RESOURCE_STATE_COPY_DEST, D3D12_RESOURCE_STATE_NON_PIXEL_SHADER_RESOURCE);
	barriers[1] = CD3DX12_RESOURCE_BARRIER::Transition(m_pathTransformedVertices.Get(), D3D12_RESOURCE_STATE_COMMON, D3D12_RESOURCE_STATE_UNORDERED_ACCESS);
	barriers[2] = CD3DX12_RESOURCE_BARRIER::Transition(m_pathVisibility.Get(), D3D12_RESOURCE_STATE_COMMON, D3D12_RESOURCE_STATE_UNORDERED_ACCESS);

	m_pathTransformComputeList->ResourceBarrier(3, barriers);

	struct STransformPushConstants
	{
		XMMATRIX m_matWorldViewProj;
		UINT m_counts[4];
		float m_dimensions[4];
	} transformPushConstants;

	constexpr UINT viewportWidth = 1024;
	constexpr UINT viewportHeight = 1024;

	transformPushConstants.m_matWorldViewProj = XMMatrixIdentity(); //matViewProj;
	transformPushConstants.m_counts[0] = static_cast<UINT>(_countof(g_pathVertices));
	transformPushConstants.m_counts[1] = 0;
	transformPushConstants.m_counts[2] = 0;
	transformPushConstants.m_counts[3] = 0;
	transformPushConstants.m_dimensions[0] = static_cast<float>(viewportWidth);
	transformPushConstants.m_dimensions[1] = static_cast<float>(viewportHeight);
	transformPushConstants.m_dimensions[2] = static_cast<float>(min(viewportWidth, viewportHeight));
	transformPushConstants.m_dimensions[3] = transformPushConstants.m_dimensions[2] * -0.5f;

	ID3D12DescriptorHeap *descriptorHeaps[] = { m_pathCBVHeap.Get() };

	m_pathTransformComputeList->SetComputeRoot32BitConstants(0, sizeof(transformPushConstants) / sizeof(int), &transformPushConstants, 0);
	m_pathTransformComputeList->SetDescriptorHeaps(1, descriptorHeaps);
	m_pathTransformComputeList->SetComputeRootDescriptorTable(1, m_pathCBVHeap->GetGPUDescriptorHandleForHeapStart());
	m_pathTransformComputeList->Dispatch(1, 1, 1);

	barriers[0] = CD3DX12_RESOURCE_BARRIER::Transition(m_pathTransformedVertices.Get(), D3D12_RESOURCE_STATE_UNORDERED_ACCESS, D3D12_RESOURCE_STATE_NON_PIXEL_SHADER_RESOURCE);
	barriers[1] = CD3DX12_RESOURCE_BARRIER::Transition(m_pathVisibility.Get(), D3D12_RESOURCE_STATE_UNORDERED_ACCESS, D3D12_RESOURCE_STATE_NON_PIXEL_SHADER_RESOURCE);

	m_pathTransformComputeList->ResourceBarrier(2, barriers);

	PIXEndEvent(m_pathTransformComputeList.Get());

	DX::ThrowIfFailed(m_pathTransformComputeList->Close());

	ID3D12CommandList *cmdLists[] = { m_pathTransformComputeList.Get() };

	m_computeQueue->ExecuteCommandLists(1, cmdLists);
	m_computeQueue->Signal(m_pathTransformFence.Get(), 1);
	m_computeQueue->Wait(m_pathTransformFence.Get(), 1);
}

void Sample3DSceneRenderer::TestCompute()
{
	if(m_testedCompute)
	{
		return;
	}

	m_testedCompute = true;

	PIXBeginEvent(m_testComputeList.Get(), 0, L"Test Compute");

	//m_testComputeList->SetPipelineState(m_testComputePipeline.Get());
	m_testComputeList->SetComputeRootSignature(m_testComputeRootSignature.Get());

	CD3DX12_RESOURCE_BARRIER barriers[3];

	barriers[0] = CD3DX12_RESOURCE_BARRIER::Transition(m_testComputeOutput.Get(), D3D12_RESOURCE_STATE_COMMON, D3D12_RESOURCE_STATE_UNORDERED_ACCESS);

	m_testComputeList->ResourceBarrier(1, barriers);

	ID3D12DescriptorHeap *descriptorHeaps[] = { m_testComputeCBVHeap.Get() };

	UINT numElements = TestComputeNumBufferItems;

	m_testComputeList->SetComputeRoot32BitConstants(0, 1, &numElements, 0);
	m_testComputeList->SetDescriptorHeaps(1, descriptorHeaps);
	m_testComputeList->SetComputeRootDescriptorTable(1, m_testComputeCBVHeap->GetGPUDescriptorHandleForHeapStart());
	//m_testComputeList->Dispatch(TestComputeNumBufferItems / 64, 1, 1);
	m_testComputeList->Dispatch(1, 1, 1);

	//barriers[0] = CD3DX12_RESOURCE_BARRIER::UAV(m_testComputeOutput.Get());
	barriers[0] = CD3DX12_RESOURCE_BARRIER::Transition(m_testComputeOutput.Get(), D3D12_RESOURCE_STATE_UNORDERED_ACCESS, D3D12_RESOURCE_STATE_COPY_SOURCE);

	m_testComputeList->ResourceBarrier(1, barriers);

	m_testComputeList->CopyResource(m_testComputeOutputDownload.Get(), m_testComputeOutput.Get());

	PIXEndEvent(m_testComputeList.Get());

	DX::ThrowIfFailed(m_testComputeList->Close());

	ID3D12CommandList *cmdLists[] = { m_testComputeList.Get() };

	m_computeQueue->ExecuteCommandLists(1, cmdLists);
	m_computeQueue->Signal(m_testComputeFence.Get(), 1);
	m_computeQueue->Wait(m_testComputeFence.Get(), 1);

	D3D12_RANGE mappedRange;

	mappedRange.Begin = 0;
	mappedRange.End = TestComputeNumBufferItems * sizeof(int);

	UINT *mappedPtr = nullptr;

	DX::ThrowIfFailed(m_testComputeOutputDownload->Map(0, &mappedRange, (void**)&mappedPtr));

	if(*mappedPtr)
	{
		OutputDebugStringW(L"Succeeded!\n");
	}
	else
	{
		OutputDebugStringW(L"Failed!\n");
	}

	m_testComputeOutputDownload->Unmap(0, nullptr);

	// If we don't sleep the reset will be caused by the next gfx command
	Sleep(10000);
}

void Sample3DSceneRenderer::TestCompute(ID3D12GraphicsCommandList *cmdList)
{
	cmdList->SetPipelineState(m_testComputePipeline.Get());
	cmdList->SetComputeRootSignature(m_testComputeRootSignature.Get());

	CD3DX12_RESOURCE_BARRIER barriers[3];

	if(!m_testedCompute)
	{
		m_testedCompute = true;

		barriers[0] = CD3DX12_RESOURCE_BARRIER::Transition(m_testComputeOutput.Get(), D3D12_RESOURCE_STATE_COMMON, D3D12_RESOURCE_STATE_UNORDERED_ACCESS);

		cmdList->ResourceBarrier(1, barriers);
	}

	ID3D12DescriptorHeap *descriptorHeaps[] = { m_testComputeCBVHeap.Get() };

	UINT numElements = TestComputeNumBufferItems;

	cmdList->SetComputeRoot32BitConstants(0, 1, &numElements, 0);
	cmdList->SetDescriptorHeaps(1, descriptorHeaps);
	cmdList->SetComputeRootDescriptorTable(1, m_testComputeCBVHeap->GetGPUDescriptorHandleForHeapStart());
	cmdList->Dispatch(TestComputeNumBufferItems / 64, 1, 1);

	barriers[0] = CD3DX12_RESOURCE_BARRIER::UAV(m_testComputeOutput.Get());

	cmdList->ResourceBarrier(1, barriers);
}

// Renders one frame using the vertex and pixel shaders.
bool Sample3DSceneRenderer::Render()
{
	// Loading is asynchronous. Only draw geometry after it's loaded.
	if (!m_loadingComplete)
	{
		return false;
	}

	/*if(!m_transformedVertices)
	{
		TransformVertices();
	}*/

	if(!m_testedCompute)
	{
		TestCompute();
	}

	DX::ThrowIfFailed(m_deviceResources->GetCommandAllocator()->Reset());

	// The command list can be reset anytime after ExecuteCommandList() is called.
	DX::ThrowIfFailed(m_commandList->Reset(m_deviceResources->GetCommandAllocator(), m_pipelineState.Get()));

	PIXBeginEvent(m_commandList.Get(), 0, L"Draw the cube");
	{
		// Set the graphics root signature and descriptor heaps to be used by this frame.
		m_commandList->SetGraphicsRootSignature(m_rootSignature.Get());
		ID3D12DescriptorHeap* ppHeaps[] = { m_cbvHeap.Get() };
		m_commandList->SetDescriptorHeaps(_countof(ppHeaps), ppHeaps);

		// Bind the current frame's constant buffer to the pipeline.
		CD3DX12_GPU_DESCRIPTOR_HANDLE gpuHandle(m_cbvHeap->GetGPUDescriptorHandleForHeapStart(), m_deviceResources->GetCurrentFrameIndex(), m_cbvDescriptorSize);
		m_commandList->SetGraphicsRootDescriptorTable(0, gpuHandle);

		// Set the viewport and scissor rectangle.
		D3D12_VIEWPORT viewport = m_deviceResources->GetScreenViewport();
		m_commandList->RSSetViewports(1, &viewport);
		m_commandList->RSSetScissorRects(1, &m_scissorRect);

		// Indicate this resource will be in use as a render target.
		CD3DX12_RESOURCE_BARRIER renderTargetResourceBarrier =
			CD3DX12_RESOURCE_BARRIER::Transition(m_deviceResources->GetRenderTarget(), D3D12_RESOURCE_STATE_PRESENT, D3D12_RESOURCE_STATE_RENDER_TARGET);
		m_commandList->ResourceBarrier(1, &renderTargetResourceBarrier);

		// Record drawing commands.
		D3D12_CPU_DESCRIPTOR_HANDLE renderTargetView = m_deviceResources->GetRenderTargetView();
		D3D12_CPU_DESCRIPTOR_HANDLE depthStencilView = m_deviceResources->GetDepthStencilView();
		m_commandList->ClearRenderTargetView(renderTargetView, DirectX::Colors::CornflowerBlue, 0, nullptr);
		m_commandList->ClearDepthStencilView(depthStencilView, D3D12_CLEAR_FLAG_DEPTH, 1.0f, 0, 0, nullptr);

		m_commandList->OMSetRenderTargets(1, &renderTargetView, false, &depthStencilView);

		m_commandList->IASetPrimitiveTopology(D3D_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
		m_commandList->IASetVertexBuffers(0, 1, &m_vertexBufferView);
		m_commandList->IASetIndexBuffer(&m_indexBufferView);
		m_commandList->DrawIndexedInstanced(36, 1, 0, 0, 0);

		// Indicate that the render target will now be used to present when the command list is done executing.
		CD3DX12_RESOURCE_BARRIER presentResourceBarrier =
			CD3DX12_RESOURCE_BARRIER::Transition(m_deviceResources->GetRenderTarget(), D3D12_RESOURCE_STATE_RENDER_TARGET, D3D12_RESOURCE_STATE_PRESENT);
		m_commandList->ResourceBarrier(1, &presentResourceBarrier);
	}
	PIXEndEvent(m_commandList.Get());

	//TestCompute(m_commandList.Get());

	DX::ThrowIfFailed(m_commandList->Close());

	// Execute the command list.
	ID3D12CommandList* ppCommandLists[] = { m_commandList.Get() };
	m_deviceResources->GetCommandQueue()->ExecuteCommandLists(_countof(ppCommandLists), ppCommandLists);

	return true;
}
